"""Common types used in :mod:`pyrcv`."""
from dataclasses import dataclass

from icontract import invariant


class PyRcvError(Exception):
    """Error in pyrcv."""


@dataclass
class RaceMetadata:
    """Specification of a single race.

    :param race_name: Unique name for this race.
    :param num_winners: How many candidates will win the race.
    :param names: List of candidate names.
    """

    race_name: str
    num_winners: int
    names: list[str]

    def __str__(self):
        cands = ",".join(self.names)
        return (
            f"race: {self.race_name}\n"
            f"num_winners: {self.num_winners}\n"
            f"candidates: {cands}"
        )


@invariant(lambda self: len(self.metadata.names) >= max(map(max, self.ballots)))
@invariant(lambda self: len(self.ballots) == len(self.votes))
@dataclass
class RaceData:
    """Voting data for a single race.

    The two main ways to use this class:

    * Each ballot corresponds to one person's vote.  Multiple ballots can be identical,
      and all the entries in votes are 1.  This usage is the most verbose and uses more
      memory, but does correspond to the typical understanding of a ballot.

    * Each ballot corresponds to a unique ordering of the candidates.  All ballots
      are unique.  The entry in :attr:`votes` corresponding to a given ballot indicates
      the number of people who voted in that ordering.  This usage is the most compact
      and uses less memory.

    As an example, the following would be identical in a 2 candidate race with 7 voters:

    .. code-block:: text

        metadata: <elided>
        ballots: [[1,2], [2], [2,1], [2,1], [2], [1], [2, 1]]
        votes: [1, 1, 1, 1, 1, 1, 1]

    .. code-block:: text

        metadata: <elided>
        ballots: [1], [2], [1, 2], [2, 1]
        votes: [1, 2, 1, 3]

    :param metadata: Details about the race.
    :param ballots: A list of candidate rankings.  Each list[int] is an ordering of a
        subset candidate indexes, with the index refering to the list of candidates in
        :attr:`metadata`.
    :param votes: A list of the same length as :attr:`ballots` which denotes the number
        of votes corresponding to each candidate ranking.  The sum of votes corresponds
        to the total number of votes cast.
    """

    metadata: RaceMetadata
    ballots: list[list[int]]
    votes: list[int]


@dataclass
class RoundResult:
    """The full results of a single-transferable voting round.

    :param count: The votes for each candidate.  Candidates are index starting at 1,
        with the index=0 reserved for exhausted ballots.
    :param elected: Candidate indices that won the election by this round.
    :param eliminated: Candidate indices that lost the election by this round.
    :param transfers: Vote counts transferred during this round.  It is a
        two-level map of ``src_cand_index -> tgt_cand_index -> vote_count``
    """

    count: list[float]
    elected: list[int]
    eliminated: list[int]
    transfers: dict[int, dict[int, float]]


@invariant(lambda self: len(set(len(r.count) for r in self.rounds)) < 2)
@invariant(
    lambda self: not self.rounds
    or (len(self.rounds[0].count) == len(self.metadata.names) + 1)
)
@dataclass
class RaceResult:
    """The list of all round results for a single race.

    :param metadata: Details about the race.
    :param rounds: List of results from each round of tabulation.
    """

    metadata: RaceMetadata
    rounds: list[RoundResult]

    def __str__(self):
        def rnd2str(round_result):
            details = []
            names_all = ["<exhausted>"] + self.metadata.names
            for idx, (name, cnt) in enumerate(zip(names_all, round_result.count)):
                name_detail = f" {name}: {cnt}"
                if idx in round_result.elected:
                    name_detail += " +"
                if idx in round_result.eliminated:
                    name_detail += " -"
                details.append(name_detail)
                # Add transfers
            return "\n".join(details)

        round_info = []
        for rnd, round_results in enumerate(self.rounds):
            round_info.extend([f"Round {rnd}:", rnd2str(round_results)])
        rounds = "\n".join(round_info)
        return f"{self.metadata}\n\n{rounds}"



"""Utilities to convert a csv format from Google Forms into pyrcv standard format."""
import re
from typing import Tuple

import numpy as np
import pandas as pd
from pandas._typing import FilePath, ReadCsvBuffer

from .types import PyRcvError, RaceData, RaceMetadata

QUESTION_PATTERN = re.compile(
    r"^(?P<question>.*?)"
    r"(\s+\((?P<num_winners>\d+)\s+winners?\))?"
    r"\s*? "
    r"\[(?P<option>.*)\]$"
)


def parse_google_form_csv(
    buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str],
) -> list[RaceData]:
    """Parses race and ballot info from Googe Form CSV results file.

    The required format is one header line, followed by one line for each ballot.
    The column headers are parsed to determine the races and candidates.  For
    example, the following represents two races, one with 3 candidates and
    one with 2 candidates:

    Mayor [Abe], Mayor [Betty], Mayor [Chris], Police Chief [Alice], Police Chief [Bob]

    Each ballot should provide a numerical ranking within each race.  A ballot cannot
    contain duplicate values within a single race.  Using the example above, some
    valid ballots are:

    1, 2, 1, 2, 3  # Can use raw numbers
    1st, 2nd, 1st, 2nd, 3rd  # Can use ordinals
    1st, 2nd,,,, 1st  # Does not rank Chris or Alice for Police Chief.
    2nd,, 1st, 2nd, 3rd  # Gap in ranking Mayoral race.  Gaps are OK.

    An example invalid ballot would be:
    1, 1, 2, 1, 3  # Duplicate ranking in Mayor race.

    :param buffer: CSV-parseable data in the format described above.
    :return: List containing an entry for each race parsed from the CSV file.
    """
    df = pd.read_csv(buffer)
    race_infos = []

    weights = df["weight"].values if "weight" in df else None

    for metadata, slice_ in parse_header(df.columns):
        goog = df.iloc[:, slice_].map(_coerce).values
        goog = np.ma.array(goog, mask=(goog == 0))

        argsort = goog.argsort(axis=1)
        mask = np.take_along_axis(goog.mask, argsort, axis=1)
        ballots = np.ma.array(argsort, mask=mask) + 1
        ballots = ballots.filled(0)
        if weights is not None:
            ballots = np.repeat(ballots, weights, axis=0)

        ballots, votes = np.unique(ballots, axis=0, return_counts=True)
        race_infos.append(RaceData(metadata, ballots.tolist(), votes.tolist()))
    return race_infos


def parse_header(header: list[str]) -> list[Tuple[RaceMetadata, slice]]:
    """Parse header row list into metadata and a column slices for each race.

    The Google Form header pattern is the race, followed by one of the options for
    that race in brackets.  Adjacent columns for the same race have the same text,
    but different options.  There also can be an optional parenthetical
    indicating the number of winners is allowed between the race and the option;
    if this parenthetical is missing, the race is assumed to be single-winner.

        :param header: Header row from CSV file
        :return: List of tuples, one for each race.  The tuple contains
            :class:`RaceMetadata` and a slice indicating the columns corresponding
            to the options for the race.

    Examples header values:

      * ``What is your favorite season? [Spring]``
      * ``City Council (4 winners) [Darth Vader]``
      * ``Mayor (1 winner) [Luke Skywalker]``
    """
    current_question = None
    current_options = []

    questions = []
    options = []
    num_winners_list = []
    starts = []
    ends = []
    for col_idx, col in enumerate(header):
        match = QUESTION_PATTERN.match(col)
        if match:
            question = match.group("question").strip()
            num_winners = match.group("num_winners") or 1
            option = match.group("option").strip()
            if question != current_question:
                if current_question is not None:
                    ends.append(col_idx)
                    options.append(current_options)
                    current_options = []
                questions.append(question)
                num_winners_list.append(int(num_winners))
                starts.append(col_idx)
                current_question = question
            current_options.append(option)
        else:
            if current_question is not None:
                current_question = None
                ends.append(col_idx)
                options.append(current_options)
                current_options = []

    if current_question is not None:
        ends.append(col_idx + 1)
        options.append(current_options)

    num_questions = len(questions)
    assert len(options) == num_questions, options
    assert len(starts) == num_questions, starts
    assert len(ends) == num_questions, ends

    return [
        (RaceMetadata(q, w, o), slice(s, e))
        for q, w, o, s, e in zip(questions, num_winners_list, options, starts, ends)
    ]


def _coerce(x):
    if isinstance(x, int):
        return x
    if isinstance(x, float):
        if np.isnan(x):
            return 0
        else:
            raise ValueError(f"Cannot use float for candidate index: {x}")
    numbers = re.findall(r"(\d+)[st|nd|rd|th]?", x)
    if not numbers or len(numbers) > 1:
        raise PyRcvError(f"Could not determine number: {x}")
    return int(numbers[0])



"""Tests for `pyrcv` package."""
import pathlib

import pytest
from click.testing import CliRunner
from numpy.testing import assert_equal

import pyrcv
from pyrcv import RaceData, RaceMetadata, RaceResult, RoundMode
from pyrcv import RoundResult as RR
from pyrcv import cli

from .utils import assert_race_equal

TESTDATA_PATH = pathlib.Path(__file__).parent / "testdata"


FAIRVOTE_BALLOTS = [
    ([1, 2, 3], 625),  # R4
    ([1, 2, 4], 125),  # R4
    ([1, 2, 5], 250),  # R4; Last is any unviable candidate
    ([1, 2, 6], 250),  # R4; Last is any unviable candidate
    ([1, 5, 3], 500),  # R3
    ([1, 5, 4], 500),  # R3
    ([1, 3], 250),
    ([2, 3, 0], 875),  # R4
    ([2, 4], 175),  # R4
    ([2, 5, 0], 350),  # R4; Last is any unviable candidate
    ([2, 6, 0], 350),  # R4; Last is any unviable candidate
    ([3], 1300),
    ([4, 0, 0], 1300),
    ([5, 2, 3], 625),  # R3, R4
    ([5, 2, 4], 125),  # R3, R4
    ([5, 2, 6], 500),  # R3, R4; Last is any unviable candidate
    ([5, 3], 100),  # R3
    ([6, 3, 0], 580),
    ([6, 4], 300),
    ([6, 2, 3], 50),  # R4
    ([6, 2, 4], 10),  # R4
    ([6, 2, 5], 40),  # R4; Last is any unviable candidate
    ([6, 5, 3], 10),  # R3
    ([6, 5, 4], 10),  # R3
]


def test_bad_rounding():
    metadata = RaceMetadata("race", num_winners=1, names=["A", "B"])
    ballots = [[2, 1], [1, 2]]
    votes = [2, 1]
    with pytest.raises(ValueError):
        pyrcv.run_rcv(RaceData(metadata, ballots, votes), round_mode="bad")


def test_2cands_1seat():
    metadata = RaceMetadata("race", num_winners=1, names=["A", "B"])
    ballots = [[2, 1], [1, 2]]
    votes = [2, 1]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(metadata, [RR([0, 1, 2], [2], [], {})])
    assert_race_equal(actual, desired)


def test_2cands_1seat_undervote():
    metadata = RaceMetadata("race", num_winners=1, names=["A", "B"])
    ballots = [[2, 0], [2, 1], [1, 2]]
    votes = [1, 1, 1]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(metadata, [RR([0, 1, 2], [2], [], {})])
    assert_race_equal(actual, desired)


def test_4cands_2seat_undervote():
    metadata = RaceMetadata("race", num_winners=2, names=["A", "B", "C", "D"])
    ballots = [[4, 0], [3, 0], [2, 4], [2, 0], [1, 4], [1, 0]]
    votes = [3, 3, 1, 1, 1, 1]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(
        metadata,
        [
            RR([0, 2, 2, 3, 3], [], [1], {1: {0: 1, 4: 1}}),
            RR([1, 0, 2, 3, 4], [4], [], {}),
            RR([1, 0, 2, 3, 4], [], [2], {2: {0: 2}}),
            RR([3, 0, 0, 3, 4], [3], [], {}),
        ],
    )
    assert_race_equal(actual, desired)


def test_3cands_2seats_1round():
    metadata = RaceMetadata("race", num_winners=2, names=["A", "B", "C"])
    ballots = [[2, 1, 3], [1, 2, 3]]
    votes = [3, 2]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(metadata, [RR([0, 2, 3, 0], [1, 2], [], {2: {3: 1}})])
    assert_race_equal(actual, desired)


def test_3cands_1seat_multiround():
    metadata = RaceMetadata("race", num_winners=1, names=["A", "B", "C"])
    ballots = [[1, 2, 3], [2, 1, 3], [3, 1, 2]]
    votes = [2, 2, 1]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(
        metadata,
        [
            RR([0, 2, 2, 1], [], [3], {3: {1: 1}}),
            RR([0, 3, 2, 0], [1], [], {}),
        ],
    )
    assert_race_equal(actual, desired)


def test_3cands_2seats_multiround():
    metadata = RaceMetadata("race", num_winners=2, names=["A", "B", "C"])
    ballots = [[1, 3, 2], [2, 1, 3], [3, 1, 2], [3, 2, 1]]
    votes = [2, 4, 1, 2]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(
        metadata,
        [
            RR([0, 2, 4, 3], [2], [], {}),
            RR([0, 2, 4, 3], [], [1], {1: {3: 2}}),
            RR([0, 0, 4, 5], [3], [], {}),
        ],
    )
    assert_race_equal(actual, desired)


def test_3cands_2seats_multiround_with_adjust():
    metadata = RaceMetadata("race", num_winners=2, names=["A", "B", "C"])
    ballots = [[1, 3, 2], [2, 1, 3], [3, 2, 1], [3, 2, 1]]
    votes = [2, 5, 1, 2]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes))
    desired = RaceResult(
        metadata,
        [
            RR([0, 2, 5, 3], [2], [], {2: {1: 1}}),
            RR([0, 3, 4, 3], [], [1], {1: {3: 3}}),
            RR([0, 0, 4, 6], [3], [], {}),
        ],
    )
    assert_race_equal(actual, desired)


def test_3cands_2seats_multiround_fractional_round():
    metadata = RaceMetadata("race", num_winners=2, names=["A", "B", "C"])
    ballots = [[1, 3, 2], [2, 1, 3], [3, 1, 2], [3, 2, 1]]
    votes = [2, 4, 1, 2]
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes), RoundMode.FRACTIONAL)
    desired = RaceResult(
        metadata,
        [
            RR([0, 2, 4, 3], [2], [], {2: {1: 0.99999}}),
            RR([0, 2.99999, 3.00001, 3], [], [1], {1: {3: 2.99999}}),
            RR([0, 0, 3.00001, 5.99999], [3], [], {}),
        ],
    )
    assert_race_equal(actual, desired)


def test_fairvote_example():
    """Example from FairVote's website.

    https://fairvote.org/archives/multi_winner_rcv_example/
    """
    metadata = RaceMetadata("race", num_winners=3, names=["A", "B", "C", "D", "E", "F"])
    ballots, votes = zip(*FAIRVOTE_BALLOTS)
    actual = pyrcv.run_rcv(RaceData(metadata, ballots, votes), RoundMode.CEILING)
    desired = RaceResult(
        metadata,
        [
            RR(
                [0, 2500, 1750, 1300, 1300, 1350, 1000],
                [1],
                [],
                {1: {2: 100, 3: 20, 5: 80}},
            ),
            RR(
                [0, 2300, 1850, 1320, 1300, 1430, 1000],
                [],
                [6],
                {6: {2: 100, 3: 580, 4: 300, 5: 20}},
            ),
            RR(
                [0, 2300, 1950, 1900, 1600, 1450, 0],
                [],
                [5],
                {5: {2: 1250, 3: 150, 4: 50}},
            ),
            RR(
                [0, 2300, 3200, 2050, 1650, 0, 0], [2], [], {2: {0: 360, 3: 450, 4: 90}}
            ),
            RR([360, 2300, 2300, 2500, 1740, 0, 0], [3], [], {3: {0: 200}}),
        ],
    )
    assert_race_equal(actual, desired)


def test_validate_and_standardize_ballots_ok():
    ballots = [[1, 0, 0], [1, 2, 0], [1, 2, 3]]
    result = pyrcv.validate_and_standardize_ballots(ballots, num_cands=3)
    assert_equal(result, [[1, 0, 0, 0], [1, 2, 0, 0], [1, 2, 3, 0]])


def test_validate_and_standardize_ballots_ragged():
    ballots = [[1, 0, 0], [1, 2], [1, 2, 3]]
    result = pyrcv.validate_and_standardize_ballots(ballots, num_cands=3)
    assert_equal(result, [[1, 0, 0, 0], [1, 2, 0, 0], [1, 2, 3, 0]])


def test_validate_and_standardize_ballots_bad_data():
    with pytest.raises(ValueError, match=r"Bad value\(s\) on ballots: \[1, 2\]"):
        ballots = [[1, 0, 0], [1, 2, -1], [1, 2, 3]]
        pyrcv.validate_and_standardize_ballots(ballots, num_cands=2)
    with pytest.raises(ValueError, match="Ballot data are not list of lists"):
        pyrcv.validate_and_standardize_ballots([1, 0, 0], num_cands=2)
    with pytest.raises(ValueError, match="Ballot data are not all integers"):
        pyrcv.validate_and_standardize_ballots([[1, 0, 0.5]], num_cands=2)
    with pytest.raises(ValueError, match=r"Ballot has duplicated entry: \[1, 0, 1\]"):
        pyrcv.validate_and_standardize_ballots([[1, 0, 1]], num_cands=2)


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    help_result = runner.invoke(cli.main, ["--help"])
    assert help_result.exit_code == 0
    assert "Show this message and exit." in help_result.output

    testdata = str(TESTDATA_PATH / "transform_testdata.csv")

    result = runner.invoke(cli.main, [testdata])
    assert result.exit_code == 0
    expected = """race: What is your favorite season?
winner(s): Autumn
"""
    assert result.output == expected

    result = runner.invoke(cli.main, [testdata, "--details"])
    assert result.exit_code == 0
    expected = """race: What is your favorite season?
num_winners: 1
candidates: Spring,Summer,Autumn,Winter

Round 0:
 <exhausted>: 0.0
 Spring: 2.0
 Summer: 3.0
 Autumn: 3.0
 Winter: 0.0 -
Round 1:
 <exhausted>: 0.0
 Spring: 2.0 -
 Summer: 3.0
 Autumn: 3.0
 Winter: 0.0
Round 2:
 <exhausted>: 1.0
 Spring: 0.0
 Summer: 3.0 -
 Autumn: 4.0
 Winter: 0.0
Round 3:
 <exhausted>: 2.0
 Spring: 0.0
 Summer: 0.0
 Autumn: 6.0 +
 Winter: 0.0
"""
    assert result.output == expected



"""Script to tabulate single-transferable vote results from a csv file.

The format required is parsed in the function
:func:`pyrcv.transform.parse_google_form_csv`.
"""
from typing import TextIO

import click

import pyrcv
from pyrcv import RoundMode


@click.command()
@click.argument(
    "csvfile",
    type=click.File(),
)
@click.option("--details", is_flag=True, help="Include round-by-round results.")
@click.option(
    "--round_mode",
    type=click.Choice(RoundMode._member_names_, case_sensitive=False),
    default=RoundMode.ADD_ONE_FLOOR.name,
    show_default=True,
    help=(
        "How to round a fractional vote threshold during tabulation. "
        "The default should be used for most cases."
    ),
)
def main(csvfile: TextIO, details: bool, round_mode: RoundMode):
    """Tabulates results of single-transferable vote elections stored in CSVFILE."""
    races = pyrcv.parse_google_form_csv(csvfile)
    round_mode = pyrcv.RoundMode[round_mode]

    for race in races:
        result = pyrcv.run_rcv(race, round_mode)

        if details:
            print(result)
        else:
            print(f"race: {race.metadata.race_name}")
            names = result.metadata.names
            # -1 because the elected indexing starts from 1.
            winners = ", ".join(
                [names[e - 1] for r in result.rounds for e in r.elected]
            )
            print(f"winner(s): {winners}")

    return 0


if __name__ == "__main__":
    main()  # pragma: no cover
